@@ -146,9 +146,6 @@ static int ap_process_http_async_connection(conn_rec *c)
             ap_set_conn_count(c->sbh, r, c->keepalives);
         }
         if ((r = ap_read_request(c))) {
-            c->keepalive = AP_CONN_UNKNOWN;
-            /* process the request if it was read without error */
-
             if (r->status == HTTP_OK) {
                 cs->state = CONN_STATE_HANDLER;
                 if (ap_extended_status) {
@@ -204,9 +201,6 @@ static int ap_process_http_sync_connection(conn_rec *c)
             keep_alive_timeout = c->base_server->keep_alive_timeout;
         }
 
-        c->keepalive = AP_CONN_UNKNOWN;
-        /* process the request if it was read without error */
-
         if (r->status == HTTP_OK) {
             if (cs)
                 cs->state = CONN_STATE_HANDLER;
